local localize = require("localize")

local function vtable_bind(class, _type, index)
    local this = ffi.cast("void***", class)
    local ffitype = ffi.typeof(_type)
    return function (...)
        return ffi.cast(ffitype, this[0][index])(this, ...)
    end
end

local WeaponSystem  = ffi.cast("void**", utils.find_pattern("client.dll", "8B 35 ? ? ? ? FF 10 0F B7 C0") + 2)[0]
local g_pMemAlloc           = ffi.cast("void***", utils.find_pattern("tier0.dll", "8B 0D ? ? ? ? 8B 11 8D") + 2)

local GetWeaponData = vtable_bind(WeaponSystem, "uintptr_t(__thiscall*)(void*, short)", 2)
local Free          = vtable_bind(g_pMemAlloc, "void(__thiscall*)(void*, void*)", 5)
local UTIL_LoadFileForMe    = ffi.cast("uint8_t*(__thiscall*)(const char*)", utils.find_pattern("client.dll", "55 8B EC 83 EC 08 56 57 8B F9 8B 0D ? ? ? ? 83"))

local function ReadFileData(path, type)
    local RawData = UTIL_LoadFileForMe(path)
    if RawData ~= nil then
        local Data = ffi.cast("uint8_t*", tonumber(ffi.cast("uintptr_t", RawData)))
        local Ret = nil
        if type == "table" then
            Ret = {}
            local Index = 0
            while true do
                table.insert(Ret, Data[Index])
                if Data[Index] == 0x00 then
                    break
                end
                Index = Index + 1
            end
        else -- Assume default type is a string
            Ret = ffi.string(Data)
        end
        Free(Data)
        return Ret
    end
    return nil
end

local Weapons = {}

for i = 0, 525 do
    local WeaponInfo = utils.get_weapon_info(i)
    if WeaponInfo then
        local IndexName = WeaponInfo.console_name:gsub("weapon_knife_", "knife_")
        Weapons[IndexName] = 
        {
            item_definition_index = i,
            weapon_info = WeaponInfo,
            localized_name = localize(ffi.string(ffi.cast("char**", GetWeaponData(i) + 0x88)[0])),
            get_texture = function (self, size)
                if not self then
                    utils.error_print("weapons lib: get_texture is a member function! Make sure you call it as such! weapons[1]:get_texture( )")
                    return nil
                end

                local FileName  = self.weapon_info.console_name:gsub("item_", ""):gsub("weapon_", "")
                local FileData  = ReadFileData(("materials/panorama/images/icons/equipment/%s.svg"):format(FileName))
                return FileData and render.create_texture_svg(FileData, size or 100) or utils.error_print("weapons lib: could not create texture")
            end
        }
    end
end

local WeaponsMT = 
setmetatable(Weapons,
{
    __index = function(self, key)
        if type(key) == "number" then
            for _, Data in pairs(Weapons) do
                if key == Data.item_definition_index then
                    return Data
                end
            end
        else
            return rawget(Weapons, key)
        end
    end,

    __call = function (table, key)
        return table[key]
    end
})

return WeaponsMT